classdef bauagent < handle
    %BAUAGENT Diese Klasse definiert die Eigenschaften des einzelnen Agenten
    %   Hier werden die Eigenschaften festgelegt, die den einzelnen Agenten
    %   ausmachen, der für den Bau der Wabenstruktur zuständig ist.
    %   Für jeden einzelnen Agenten wird ein Objekt dieser Klasse erzeugt.
    
    properties
        Position % Vektor, welcher die Position des Agenten in 
                 % X-Y(-Z)-Koordinaten enthält. Je nach Umgebung 
                 % zwei- oder dreidimensional.
        Ausrichtung % Blickrichtung des Agenten im Gradmaß (0 bis 360)
                    % Null entspricht Blick waagerecht nach rechts
        RuleSet 
                 % Enthält die Regeln, nach denen gebaut oder nicht
                 % gebaut wird. Genaue Codierung folgt.
        BuildCounter % Zählt die Anzahl der gesetzten Bausteine.
        Abstand % Ein Abstandsvektor, der die absoluten Abstandswerte zu
                % allen Zellenmittelpunkten enthält
        Speed = 1; % relative Geschwindigkeit, mit der sich der Agent bewegt
        
        Markierung % graphisches Objekt zur Positionsmarkierung
        Index % enthält die laufende Nummer des Agenten
        Farbe % die Farbe zur Markierung
    end
    
    methods
        function obj = bauagent(InitialPosition, indexNummer) % Konstruktor
            global Grenzen;
            obj.Ausrichtung = 0;
            obj.RuleSet = ['BBB***1'; 'B*BB**1'; 'B**BB*1'; ...
                           'B***BB1'; '*BBB**1'; '*B*BB*1'; ...
                           '*B**BB1'; '**BBB*1'; '**BB*B1'; ...
                           '***BBB1'; '*BB***2'; '****BB2'; ...
                           '***B**3'; 'B*****3' ];
            
            % obj.RuleSet = ['B*****1'; '***B**1'; '*B****1'; ...
            %               '****B*1'; '**B***1'; '*****B1' ]; % Regelwerk zum Testen
            
            obj.Index = indexNummer; % übergibt den Index aus dem Vektor
                                     % von Agenten als Eigenschaft
            
            if (mod(obj.Index, 7) == 1)
                obj.Farbe = 'blue';
            elseif (mod(obj.Index, 7) == 2)
                obj.Farbe = 'cyan';
            elseif (mod(obj.Index, 7) == 3)
                obj.Farbe = 'green';
            elseif (mod(obj.Index, 7) == 4)
                obj.Farbe = 'white';
            elseif (mod(obj.Index, 7) == 5)
                obj.Farbe = 'black';
            elseif (mod(obj.Index, 7) == 6)
                obj.Farbe = 'yellow';
            else
                obj.Farbe = 'magenta';
            end
            
            obj.BuildCounter = 0; % Anzahl der gebauten Bausteine auf Null
            
            if nargin > 0
                obj.Position = InitialPosition;
            else
                obj.Position = rand(1,2);
                obj.Position(1,1) = obj.Position(1,1) * Grenzen(1); % Setzt die Position des Agenten auf einen
                obj.Position(1,2) = obj.Position(1,2) * Grenzen(2); % beliebigen Punkt in den Grenzen des Gitters
            end
            
            obj.Markierung = line(obj.Position(1), obj.Position(2), ...
                                    1.1, 'Marker','.', 'MarkerFaceColor',obj.Farbe, ...
                                    'Color',obj.Farbe );
                         % aktuelle Position des Agenten wird graphisch
                         % markiert
            % set(obj.Markierung, 'XDataSource','agent.Position(1)');
            % set(obj.Markierung, 'YDataSource','agent.Position(2)');
                        % Verknüpfung des Grafikobjekts mit der Position
        end
        
        function obj = movebuild(agent) % Diese Funktion wird einmal pro Zeitschritt für jeden Agenten aufgerufen
            % Die Umgebung wird verarbeitet und nach den Regeln des Agenten
            % entschieden, ob der Agent an der aktuellen Position eine Wabe baut
            % oder sich weiterbewegt.
            global Wabengitter;
            global Knotenzahl;
            global Grenzen;
            global raumplot;
            
            agent.Ausrichtung = mod(agent.Ausrichtung + rand*360, 360); % dreht den Agenten in eine zufällige Richtung
            
            Zelle = agent.indexPosition;      % Index der Zelle, auf dessen Fläche sich der Agent befindet
            Nachbarn = agent.nachbarn(Zelle); % erzeugt Liste der Indizes der den Agenten unmittelbar
                                              % umgebenden Zellen
            
            % refreshdata(agent.Markierung); % Markierung wird neu gezeichnet
            
            
            for nachbarZelle=1:6
                builtFlag = 0;
                validFlag = 1; % Initialisierung
                
                bauPosition = Nachbarn(nachbarZelle); 
                % setze potentielle Position für neuen Baustein auf einen
                % der sechs Nachbarzellen der aktuellen Position
                
                if ( (bauPosition < 1) || (bauPosition > Knotenzahl) || ... % Grenzen des Gitters
                     (Wabengitter(4,bauPosition) ~= 0) || ... % Bebauung der Zelle
                     (Wabengitter(3,bauPosition) == 0) ) % und ist die Wabe noch im Gitter?
                    validFlag = 0;  
                                   
                end
                
                bauNachbarn = agent.nachbarn(bauPosition); % erzeugt Liste der unmittelbaren Nachbarn der
                                                           % potentiellen
                                                           % Bauposition
                
                for ruleNumber=1:size(agent.RuleSet, 1) % jede Regel wird geprüft
                     
                    
                    if (validFlag == 0)
                        break; % wenn sich die Position außerhalb des
                               % Gitters befindet oder schon bebaut ist,
                               % wird die nächste Position geprüft
                    end
                    
                    validFlag = 1;
                    
                    for ruleCellNumber=1:6 % jede Position der Regel einzeln geprüft
                        if ( (validFlag == 0) || ...
                           (bauNachbarn(ruleCellNumber) > Knotenzahl) || ...
                           (bauNachbarn(ruleCellNumber) < 1) )
                            break; % wenn schon die letzte Stelle der Regel ungültig war,
                                   % braucht diese nicht mehr getestet zu
                                   % werden - wenn die Position außerhalb
                                   % des Gitters liegt, auch nicht
                        end
                        
                        if strcmp(agent.RuleSet(ruleNumber,ruleCellNumber), 'B')
                            if (strcmp(num2str(Wabengitter(4,bauNachbarn(ruleCellNumber))),'1') ...
                             || strcmp(num2str(Wabengitter(4,bauNachbarn(ruleCellNumber))),'2') ...
                             || strcmp(num2str(Wabengitter(4,bauNachbarn(ruleCellNumber))),'3') ...
                             || strcmp(num2str(Wabengitter(4,bauNachbarn(ruleCellNumber))),'4') ...
                             || strcmp(num2str(Wabengitter(4,bauNachbarn(ruleCellNumber))),'5') ...
                             || strcmp(num2str(Wabengitter(4,bauNachbarn(ruleCellNumber))),'6') ...
                             || strcmp(num2str(Wabengitter(4,bauNachbarn(ruleCellNumber))),'7') )
                                 validFlag = 1; % wenn die Regel an dieser Stelle 'B' enthält,
                                 % wird validFlag 'true' gesetzt, falls ein
                                 % beliebiger Baustein vorhanden ist.
                            else
                                validFlag = 0;
                                
                            end
                        elseif strcmp(agent.RuleSet(ruleNumber,ruleCellNumber), '*')
                            validFlag = 1; % enthält die Regel an der betreffenden Stelle ein '*'
                                           % so ist der Zustand der Zelle
                                           % egal.
                        else
                            validFlag = strcmp(num2str(Wabengitter(4,bauNachbarn(ruleCellNumber))), ...
                                           agent.RuleSet(ruleNumber,ruleCellNumber));
                                       % in allen anderen Fällen (genaue
                                       % Spezifikation in der Regel) muss
                                       % der Zelleninhalt exakt mit der
                                       % Regelposition übereinstimmen.
                        end
                    
                    end
                    
                    if (validFlag == 1)
                        Wabengitter(4,bauPosition) = str2double(agent.RuleSet(ruleNumber,7));
                                % hier wird gebaut
                        agent.BuildCounter = agent.BuildCounter + 1;
                                % Counter erhöhen
                        
                        builtFlag = 1; % zeigt an, dass in diesem Zeitschritt gebaut wurde
                        
                        plotwabe(bauPosition, raumplot, Wabengitter(4,bauPosition)); 
                                % die gebaute Wabe wird markiert
                                % Parameter:
                                % bauPosition = Index der zu bebauenden
                                % Zelle
                                % raumplot = Handle des zu benutzenden
                                % Achsenobjekts
                                % Wabengitter(...) = Typ des Bausteins
                                % (definiert die Farbe)
                        break;
                            % wenn gebaut wurde, werden die restlichen Regeln
                            % ignoriert
                    end
                    
                    validFlag = 1; % sonst wird die nächste Regel nicht angewendet
                end
                
                if true(builtFlag)
                    builtFlag = 0;
                    break; % jeder Agent soll höchstens einmal pro Zeitschritt bauen
                end
            end
               
            
            agent.Position(1) = agent.Position(1) + cosd(agent.Ausrichtung * agent.Speed); % Bewegung in x-Richtung
            agent.Position(2) = agent.Position(2) + sind(agent.Ausrichtung * agent.Speed); % Bewegung in y-Richtung
            
            zelle = agent.indexPosition;
            if ( (Wabengitter(4,zelle) == 0) || ... % wenn Position unbebaut ist...
                 (agent.Position(1) > Grenzen(1)) || ...
                 (agent.Position(1) < 0) ||...
                 (agent.Position(2) > Grenzen(2)) || ...
                 (agent.Position(2) < 0) ) % .. oder außerhalb des Gitters liegt
                         % ...dann mache Positionsänderung rückgängig
                agent.Position(1) = agent.Position(1) + ...
                    cosd( mod((agent.Ausrichtung + 180),360) * agent.Speed );
                agent.Position(2) = agent.Position(2) + ...
                    sind( mod((agent.Ausrichtung + 180),360) * agent.Speed );
            end
                
            delete(agent.Markierung); % alte Markierung wird gelöscht
            agent.Markierung = line(agent.Position(1), agent.Position(2), ...
                                    1.1, 'Marker','.', 'MarkerFaceColor',agent.Farbe, ...
                                    'Color',agent.Farbe, 'Parent',raumplot );
                                    % und neue Markierung gesetzt
            drawnow;
            
            
        end
        
        function liste = nachbarn(obj, mittelpunkt)
            % Gibt für einen gegebenen Knoten im Wabengitter die Liste der
            % direkten Nachbarn zurück - Reihenfolge vom oberen Nachbarn 
            % beginnend im Uhrzeigersinn
            global DX;
            
            liste = [ mittelpunkt+2*DX, ...
                     mittelpunkt+DX+1+mod(idivide(mittelpunkt,uint32(DX)),2), ...
                     mittelpunkt-DX+1+mod(idivide(mittelpunkt,uint32(DX)),2), ...
                     mittelpunkt-2*DX, ...
                     mittelpunkt-DX-2+mod(idivide(mittelpunkt,uint32(DX)),2), ...
                     mittelpunkt+DX-2+mod(idivide(mittelpunkt,uint32(DX)),2) ];
        end
        
        function indexNummer = indexPosition(agent)
            % Gibt für jeden Agenten den Index des Mittelpunkts derjenigen
            % Wabe zurück, auf deren Fläche sich der Agent nach seinen
            % Koordinaten befindet
            global Wabengitter;
            global MittelPunkte;
            
            agent.Abstand = sqrt((Wabengitter(1,MittelPunkte) - (agent.Position(1,1))).^2 ...
                               + (Wabengitter(2,MittelPunkte) - (agent.Position(1,2))).^2 );
                         % erzeugt einen Vektor mit den euklidischen Abständen des Agenten
                         % von allen Wabenmittelpunkten: 
                         % D(X,Y) = sqrt [ (x1-y1)^2 + (x2-y2)^2 ]
          
            [abstand, I] = min(agent.Abstand); % gibt den Abstand zum nächstgelegenen
                       % Wabenmittelpunkt (abstand) und den dazugehörigen
                       % Index im Vektor Abstand (I) zurück
            indexNummer = MittelPunkte(I); % der nächste Mittelpunkt ist der
                        % mit Index I im Vektor Abstand, also ist sein
                        % Index im Array Wabengitter der Wert an Stelle i
                        % im Vektor Abstand
        end
        
    end

end

